import dataclasses
import os
from types import ModuleType
from typing import *

from process_bigraph import ProcessTypes

from bsp.utils.base_utils import dynamic_simulator_install
from bsp.data_model.base import Implementation, Type


@dataclasses.dataclass
class ImplementationRegistry:
    id: str
    implementation: ProcessTypes
    primary: bool


class SimulatorDependency(str):
    pass


class Registrar(object):
    registries: List[ImplementationRegistry]
    core: ProcessTypes
    registered_addresses: List[str]
    implementation_dependencies: Dict[str, List[SimulatorDependency]]

    def __init__(self, core: ProcessTypes = None):
        self.core = core or ProcessTypes()
        self.registries = []

        default_reg = ImplementationRegistry(
            id="default",
            implementation=ProcessTypes(),
            primary=True
        )
        self.add_registry(default_reg)
        self.set_primary(default_reg.id)
        self.core.type_registry = self.core.types()
        self.initial_registration_complete = False
        self.implementation_dependencies = {}

    @property
    def registered_addresses(self) -> List[str]:
        return list(self.core.process_registry.registry.keys())

    def add_registry(self, registry: ImplementationRegistry):
        if registry.primary:
            for registry in self.registries:
                if registry.primary is not None:
                    registry.primary = False

            self.core = registry.implementation

        self.registries.append(registry)

    def set_primary(self, registry_id: str):
        for registry in self.registries:
            # take away existing primary
            if registry.primary:
                registry.primary = False

            # set ref as primary
            if registry.id == registry_id:
                registry.primary = True
                self.core = registry.implementation

    def register_type(self, type_id: str, type_schema: Dict):
        self.core.register_types({type_id: type_schema})

    def register_process(self, address: str, implementation: object, verbose=False) -> None:
        type_registry = self.core
        type_registry.process_registry.register(address, implementation)
        if verbose:
            print(f"Successfully registered {implementation} to address: {address}")

    def register_module(self, implementation: Implementation, verbose=False, attempt_install=False) -> None:
        library, module_name, class_name = implementation.location.rsplit('.', 3)
        try:
            # library = 'steps' if 'process' not in path else 'processes'
            import_statement = f'bsp.{library}.{module_name}'
            module = __import__(
                 import_statement, fromlist=[class_name])
            bigraph_class = getattr(module, class_name)
            self.core.process_registry.register(implementation.address, bigraph_class)
            if verbose:
                print(f"Successfully registered {bigraph_class} to {implementation.address}")
        except Exception as e:
            if verbose:
                print(f"Cannot register {class_name}. Error:\n**\n{e}\n**")
            if attempt_install:
                dynamic_simulator_install(simulators=[library])

    def register_initial_modules(
            self,
            items_to_register: List[Implementation],
            package: str = "bsp",
            verbose=False,
            attempt_install=False
    ) -> None:
        if not self.initial_registration_complete:
            for implementation in items_to_register:
                self.register_module(implementation=implementation, verbose=verbose, attempt_install=attempt_install)
                process_deps = [SimulatorDependency(dep) for dep in implementation.dependencies]
                self.implementation_dependencies[implementation.address] = process_deps
            self.initial_registration_complete = True

    def register_type_module(self, module: ModuleType, verbose=False) -> None:
        for schema_name in module.__all__:
            schema = getattr(module, schema_name)
            self.register_type(schema_name, schema)
            print(f'Registered module {schema_name}') if verbose else None

    def register_initial_types(self, config: ModuleType, types: ModuleType, verbose=False) -> None:
        for module in [config, types]:
            self.register_type_module(module, verbose=verbose)



"""
Base data model relating to utils, files, protocols, etc.
"""


from dataclasses import dataclass, asdict, Field, field
from types import FunctionType
from typing import List, Dict, Union

from pydantic import ConfigDict, BaseModel as _BaseModel


class BaseModel(_BaseModel):
    """Base Pydantic Model with custom app configuration"""
    model_config = ConfigDict(arbitrary_types_allowed=True)


@dataclass
class BaseClass:
    """Base Python Dataclass multipurpose class with custom app configuration."""
    def to_dict(self):
        return asdict(self)


@dataclass
class FilePath(BaseClass):
    name: str
    path: str


# --- INSTALLATION
@dataclass
class DependencyFile(BaseClass):
    name: str
    hash: str


@dataclass
class InstallationDependency(BaseClass):
    name: str
    version: str
    markers: str  # ie: "markers": "platform_system == \"Windows\""
    # files: List[DependencyFile] = Field(default=[])


@dataclass
class Simulator(BaseClass):
    name: str  # name installed by pip
    version: str
    deps: List[InstallationDependency]


@dataclass
class Implementation(BaseClass):
    address: str
    location: str
    dependencies: List[str]


@dataclass
class Type(BaseClass):
    address: str
    schema: Dict[str, Union[str, FunctionType]]




# quantum utils
import numpy as np
from qiskit.quantum_info import SparsePauliOp, Pauli
from qiskit.result import QuasiDistribution


def bitfield(n, L):
    result = np.binary_repr(n, L)
    return [int(digit) for digit in result]


def get_operator(weight_matrix, num_nodes: int):
    r"""Generate Hamiltonian for the graph partitioning
    Notes:
        Goals:
            1 Separate the vertices into two set of the same size.
            2 Make sure the number of edges between the two set is minimized.
        Hamiltonian:
            H = H_A + H_B
            H_A = sum\_{(i,j)\in E}{(1-ZiZj)/2}
            H_B = (sum_{i}{Zi})^2 = sum_{i}{Zi^2}+sum_{i!=j}{ZiZj}
            H_A is for achieving goal 2 and H_B is for achieving goal 1.
    Args:
        weight_matrix: Adjacency matrix.
        num_nodes: N vars at iteration k.
    Returns:
        Operator for the Hamiltonian
        A constant shift for the obj function.
    """
    num_nodes = len(weight_matrix)
    pauli_list = []
    coeffs = []
    shift = 0

    for i in range(num_nodes):
        for j in range(i):
            if weight_matrix[i, j] != 0:
                x_p = np.zeros(num_nodes, dtype=bool)
                z_p = np.zeros(num_nodes, dtype=bool)
                z_p[i] = True
                z_p[j] = True
                pauli_list.append(Pauli((z_p, x_p)))
                coeffs.append(-0.5)
                shift += 0.5

    for i in range(num_nodes):
        for j in range(num_nodes):
            if i != j:
                x_p = np.zeros(num_nodes, dtype=bool)
                z_p = np.zeros(num_nodes, dtype=bool)
                z_p[i] = True
                z_p[j] = True
                pauli_list.append(Pauli((z_p, x_p)))
                coeffs.append(1.0)
            else:
                shift += 1

    return SparsePauliOp(pauli_list, coeffs=coeffs), shift


def sample_most_likely(state_vector):
    """Compute the most likely binary string from state vector.
    Args:
        state_vector: State vector or quasi-distribution.

    Returns:
        Binary string as an array of ints.
    """
    if isinstance(state_vector, QuasiDistribution):
        values = list(state_vector.values())
    else:
        values = state_vector
    n = int(np.log2(len(values)))
    k = np.argmax(np.abs(values))
    x = bitfield(k, n)
    x.reverse()
    return np.asarray(x)



"""
Tellurium Process
"""


import numpy as np
import tellurium as te
from process_bigraph import Process, Composite, pf, Step

from bsp.data_model.sed import SedModel as SEDModelType

# from biosimulators_processes import CORE
# from biosimulators_processes.processes.utc_process import SbmlUniformTimeCourse


# class UtcTellurium(SbmlUniformTimeCourse):
#     def __init__(self,
#                  config=None,
#                  core=CORE,
#                  time_config: dict = None,
#                  model_source: str = None,
#                  sed_model_config: dict = None):
#         super().__init__(config, core, time_config, model_source, sed_model_config)
#         self.floating_species_list = self.simulator.getFloatingSpeciesIds()
#         self.boundary_species_list = self.simulator.getBoundarySpeciesIds()
#         self.floating_species_initial = self.simulator.getFloatingSpeciesConcentrations()
#         self.boundary_species_initial = self.simulator.getBoundarySpeciesConcentrations()
#
#         # Get the list of parameters and their values
#         self.model_parameters_list = self.simulator.getGlobalParameterIds()
#         self.model_parameter_values = self.simulator.getGlobalParameterValues()
#
#         # Get a list of reactions
#         self.reaction_list = self.simulator.getReactionIds()
#
#     def _get_initial_state_params(self):
#         floating_species_dict = dict(zip(self.floating_species_list, self.floating_species_initial))
#         # boundary_species_dict = dict(zip(self.boundary_species_list, self.boundary_species_initial))
#         model_parameters_dict = dict(zip(self.model_parameters_list, self.model_parameter_values))
#         return {
#             'time': [0.0],
#             self.species_context_key: floating_species_dict,
#             'model_parameters': model_parameters_dict,
#             'reactions': self.reaction_list
#         }
#
#     def plot_results(self):
#         # return self.simulator.plot()
#         return super().plot_results(simulator_name='Tellurium')
#
#     def _load_simulator(self, model_fp: str, **kwargs):
#         return te.loadSBMLModel(model_fp)
#
#     def _get_reactions(self) -> list[str]:
#         return self.simulator.getReactionIds()
#
#     def _get_floating_species(self) -> list[str]:
#         return self.simulator.getFloatingSpeciesIds()
#
#     def _get_model_parameters(self) -> list[str]:
#         return self.simulator.getGlobalParameterIds()
#
#     def _generate_results(self, inputs=None):
#         x = inputs or self.initial_state()
#         # TODO: set vals if inputs here.
#         output_start = self.output_start_time + 1 if self.output_start_time < 1 else self.output_start_time
#         self.simulator.simulate(start=self.initial_time, end=output_start)
#         s = self.simulator.simulate(start=output_start, end=self.duration, steps=self.num_steps)
#         outputs = {'floating_species': {}}
#         for i, row in enumerate(s.transpose()):
#             if i < 1:
#                 outputs['time'] = row
#             else:
#                 for spec_index, name in enumerate(self.floating_species_list):
#                     outputs['floating_species'][self.output_keys[spec_index]] = row
#         return outputs


class TelluriumStep(Step):

    config_schema = {
        'sbml_model_path': 'string',
        'antimony_string': 'string',
    }

    def __init__(self, config=None, core=None):
        super().__init__(config, core)

        # initialize a tellurium(roadrunner) simulation object. Load the model in using either sbml(default) or antimony
        if self.config.get('antimony_string') and not self.config.get('sbml_model_path'):
            self.simulator = te.loada(self.config['antimony_string'])
        elif self.config.get('sbml_model_path') and not self.config.get('antimony_string'):
            self.simulator: te.roadrunner.extended_roadrunner.ExtendedRoadRunner = te.loadSBMLModel(self.config['sbml_model_path'])
        else:
            raise Exception('the config requires either an "antimony_string" or an "sbml_model_path"')

        self.input_ports = [
            'floating_species',
            'boundary_species',
            'model_parameters'
        ]

        self.output_ports = [
            'floating_species',
        ]

        # Get the species (floating and boundary)
        self.floating_species_list = self.simulator.getFloatingSpeciesIds()
        self.boundary_species_list = self.simulator.getBoundarySpeciesIds()
        self.floating_species_initial = self.simulator.getFloatingSpeciesConcentrations()
        self.boundary_species_initial = self.simulator.getBoundarySpeciesConcentrations()

        # Get the list of parameters and their values
        self.model_parameters_list = self.simulator.getGlobalParameterIds()
        self.model_parameter_values = self.simulator.getGlobalParameterValues()

        # Get a list of reactions
        self.reaction_list = self.simulator.getReactionIds()

    # TODO -- is initial state even working for steps?
    def initial_state(self, config=None):
        return {
            'inputs': {
                'time': 0,
            },
        }

    def inputs(self):
        return {
            'time': 'float',
            'run_time': 'float',
        }

    def outputs(self):
        return {
            'results': {
                '_type': 'numpy_array',
                '_apply': 'set'
            }  # This is a roadrunner._roadrunner.NamedArray
        }

    def update(self, inputs):
        results = self.simulator.simulate(
            inputs['time'],
            inputs['run_time'],
            10
        )  # TODO -- adjust the number of saves teps
        return {
            'results': results}


class TelluriumProcess(Process):
    config_schema = {
        'record_history': 'maybe[boolean]',  # TODO -- do we have this type?
        'model': SEDModelType().to_dict(),
        'method': {
            '_default': 'cvode',
            '_type': 'string'
        },
        'species_context': {
            '_default': 'concentrations',
            '_type': 'string'
        },
        'num_steps': 'maybe[integer]'
    }

    simulator: te.roadrunner.extended_roadrunner.ExtendedRoadRunner

    def __init__(self, config=None, core=None):
        super().__init__(config, core)

        model_source = self.config['model']['model_source']
        if '/' in model_source:
            self.simulator = te.loadSBMLModel(model_source)
        else:
            if 'model' in model_source:  # TODO: find a better way to do this
                self.simulator = te.loada(model_source)
            else:
                raise Exception('the config requires either an "antimony_string" or an "sbml_model_path"')

        # handle context type (concentrations for deterministic by default)
        context_type = self.config['species_context']
        self.species_context_key = f'floating_species_{context_type}'
        self.use_counts = 'concentrations' in context_type

        # TODO -- make this configurable.
        self.input_ports = [
            self.species_context_key,
            'boundary_species',
            'model_parameters'
            'time']

        self.output_ports = [
            self.species_context_key,
            'time']

        # Get the species (floating and boundary)
        self.floating_species_list = self.simulator.getFloatingSpeciesIds()
        self.boundary_species_list = self.simulator.getBoundarySpeciesIds()
        self.floating_species_initial = self.simulator.getFloatingSpeciesConcentrations()
        self.boundary_species_initial = self.simulator.getBoundarySpeciesConcentrations()

        # Get the list of parameters and their values
        self.model_parameters_list = self.simulator.getGlobalParameterIds()
        self.model_parameter_values = self.simulator.getGlobalParameterValues()

        # Get a list of reactions
        self.reaction_list = self.simulator.getReactionIds()

    def initial_state(self, config=None):
        floating_species_dict = dict(zip(self.floating_species_list, self.floating_species_initial))
        boundary_species_dict = dict(zip(self.boundary_species_list, self.boundary_species_initial))
        model_parameters_dict = dict(zip(self.model_parameters_list, self.model_parameter_values))
        return {
            'time': 0.0,
            self.species_context_key: floating_species_dict,
            # 'boundary_species': boundary_species_dict,
            'model_parameters': model_parameters_dict
        }

    def inputs(self):
        float_set = {'_type': 'float', '_apply': 'set'}
        return {
            'time': 'float',
            # 'run_time': 'float',
            self.species_context_key: {
                species_id: float_set for species_id in self.floating_species_list},
            # 'boundary_species': {
                # species_id: float_set for species_id in self.boundary_species_list},
            'model_parameters': {
                param_id: float_set for param_id in self.model_parameters_list},
            'reactions': {
                reaction_id: float_set for reaction_id in self.reaction_list},
        }

    def outputs(self):
        float_set = {'_type': 'float', '_apply': 'set'}
        return {
            self.species_context_key: {
                species_id: float_set for species_id in self.floating_species_list},
            'time': 'float'
        }

    def update(self, inputs, interval):

        # set tellurium values according to what is passed in states
        for port_id, values in inputs.items():
            if port_id in self.input_ports:  # only update from input ports
                for cat_id, value in values.items():
                    self.simulator.setValue(cat_id, value)

        for cat_id, value in inputs[self.species_context_key].items():
            self.simulator.setValue(cat_id, value)

        # run the simulation
        result = self.simulator.oneStep(inputs['time'], interval)

        # extract the results and convert to update
        update = {
            'time': interval,  # new_time,
            self.species_context_key: {
                mol_id: float(self.simulator.getValue(mol_id))
                for mol_id in self.floating_species_list
            }
        }

        """for port_id, values in inputs.items():
            if port_id in self.output_ports:
                update[port_id] = {}
                for cat_id in values.keys():
                    update[port_id][cat_id] = self.simulator.getValue(cat_id)"""
        return update


def test_process(core=None):
    # 1. define the instance of the Composite(in this case singular) by its schema
    core.process_registry.register('biosimulators_processes.processes.tellurium_process.TelluriumProcess', TelluriumProcess)
    instance = {
        'tellurium': {
            '_type': 'process',
            'address': 'local:!biosimulators_processes.processes.tellurium_process.TelluriumProcess', # using a local toy process
            'config': {
                'sbml_model_path': 'biosimulators_processes/model_files/BIOMD0000000061_url.xml',
            },
            'inputs': {
                'time': ['start_time_store'],
                'run_time': ['run_time_store'],
                'floating_species': ['floating_species_store'],
                'boundary_species': ['boundary_species_store'],
                'model_parameters': ['model_parameters_store'],
                'reactions': ['reactions_store'],
                # 'interval': ['interval_store'],
            },
            'outputs': {
                'results': ['results_store'],
            }
        }
    }

    # 2. make the composite
    workflow = Composite({
        'state': instance
    })

    # 3. run
    # update = workflow.run(10)

    # 4. gather results
    # results = workflow.gather_results()
    # print(f'RESULTS: {pf(results)}')



from typing import Dict, Union, List, Tuple
from types import FunctionType
import os

import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from basico import biomodels, load_model_from_string
from process_bigraph import Composite, pf, pp, ProcessTypes
import nbformat


def check_ode_kisao_term(term: str):
    """Check that a term is representative an ODE algorithm."""
    from kisao import utils
    ode_ids = [term.id for term in list(utils.get_ode_algorithms())]
    for i, id in enumerate(ode_ids):
        is_term = id.endswith(term)
        if is_term:
            return True
        else:
            pass
    return False


def plot_utc_outputs(data: dict, t: Union[np.array, List[float]], simulator: str = None, x=None, y=None) -> None:
    """Plot ODE simulation observables with Seaborn."""
    plt.figure(figsize=(20, 8))
    species_data = data.get('floating_species') or data.get('floating_species_concentrations')
    for spec_name, spec_data in species_data.items():
        sns.lineplot(x=t, y=spec_data, label=spec_name)

    plt.xlabel('Time')
    plt.ylabel('Concentration')
    plt.title(f'Species Concentrations over Time with {simulator or "UTC Simulator"}')
    plt.legend()
    plt.grid(True)
    plt.show()


def plot_ode_output_data(data: dict, simulator_name: str, sample_size: int = None) -> None:
    """
    Args:
        data: outermost keys are 'time' and 'floating_species'

    Returns:

    """
    time: np.ndarray = data.get('results', data['time'])
    plt.figure(figsize=(20, 8))
    for name, output in data.get('floating_species', 'floating_species_concentrations').items():
        x = np.array(time)
        y = output
        plt.plot(x, y, label=name)

    plt.xlabel('Time')
    plt.ylabel('Concentration')
    plt.title(f'Species Concentrations Over Time with {simulator_name}')
    plt.legend()
    plt.grid(True)
    # plt.xlim([0, time[-1]])  # Set x-axis limits from 0 to the last time value
    # plt.ylim([min([list(v.values())[0] for v in spec_outputs]), max([list(v.values())[0] for v in spec_outputs])])  # Adjust y-axis to fit all data
    plt.show()


def create_ode_step_instance(simulator_name: str, **step_kwargs):
    module_name = 'ode_simulation'
    import_statement = 'biosimulators_processes.steps.ode_simulation'
    class_name = f'{simulator_name.replace(simulator_name[0], simulator_name[0].upper())}Step'
    module = __import__(
        import_statement, fromlist=[class_name])
    bigraph_class = getattr(module, class_name)
    return bigraph_class(**step_kwargs)


def calc_num_steps(dur, step_size) -> int:
    return int(dur / step_size)


def calc_duration(n_steps, step_size) -> int:
    return int(n_steps * step_size)


def calc_step_size(dur, n_steps) -> float:
    return dur / n_steps


def register_module(items_to_register: List[Tuple[str, str]], core: ProcessTypes, verbose=False) -> None:
    for process_name, path in items_to_register:
        module_name, class_name = path.rsplit('.', 1)
        try:
            library = 'steps' if 'process' not in path else 'processes'
            import_statement = f'biosimulators_processes.{library}.{module_name}'

            module = __import__(
                 import_statement, fromlist=[class_name])

            # module = importlib.import_module(import_statement)

            # Get the class from the module
            bigraph_class = getattr(module, class_name)

            # Register the process
            core.process_registry.register(process_name, bigraph_class)
        except Exception as e:
            print(f"Cannot register {class_name}. Error:\n**\n{e}\n**")
            continue


def prepare_single_ode_process_document(
        process_id: str,
        simulator_address_name: str,
        sbml_model_fp: str,
        add_emitter=True
        ) -> Dict:
    """
        * `simulator_name` must correspond to an existing process implementation.
        # TODO: Make this more flexible by providing a default to biosimulators1.0 for tools not yet implemented.
    """
    species_context = 'concentrations'
    species_port_name = f'floating_species_{species_context}'
    species_store = [f'{process_id}_{species_port_name}_store']
    document = {
        process_id: {
            '_type': 'process',
            'address': f'local:{simulator_address_name}',
            'config': {
                'model': {
                    'model_source': sbml_model_fp}},
            'inputs': {
                species_port_name: species_store,
                'model_parameters': ['model_parameters_store'],
                'time': ['time_store'],
                'reactions': ['reactions_store']},
            'outputs': {
                species_port_name: species_store,
                'time': ['time_store']}}}

    if add_emitter:
        document['emitter'] = {
            '_type': 'step',
            'address': 'local:ram-emitter',
            'config': {
                'emit': {
                    species_port_name: 'tree[float]',
                    'time': 'float'}},
            'inputs': {
                species_port_name: species_store,
                'time': ['time_store']}}

    return document


def prepare_single_copasi_process_schema(
        process_name: str,
        biomodel_id: str = None,
        sbml_model_fp: str = None,
        method='lsoda',
        species_context='concentrations',
        add_emitter=True
        ) -> Dict:
    species_port_name = f'floating_species_{species_context}'
    species_store = [f'floating_species_{species_context}_store']
    model_source = biomodel_id or sbml_model_fp
    assert model_source, 'You must pass either a biomodel id or sbml model path source.'
    document = {
        process_name: {
            '_type': 'process',
            'address': 'local:copasi',
            'config': {
                'model': {
                    'model_source': model_source},
                'method': method,
                'species_context': species_context},
            'inputs': {
                species_port_name: species_store,
                'model_parameters': ['model_parameters_store'],
                'time': ['time_store'],
                'reactions': ['reactions_store']},
            'outputs': {
                species_port_name: species_store,
                'time': ['time_store']}}}

    if add_emitter:
        document['emitter'] = {
            'emitter': {
                '_type': 'step',
                'address': 'local:ram-emitter',
                'config': {
                    'emit': {
                        species_port_name: 'tree[float]',
                        'time': 'float'}},
                'inputs': {
                    species_port_name: species_store,
                    'time': ['time_store']}}}

    return document


def run_composition(instance, duration: int = 10) -> dict:
    workflow = Composite(
        config={
            'state': instance
        }
    )
    workflow.run(duration)
    return workflow.gather_results()


def create_class_from_dict(instance_dict: Dict, class_name="DynamicClass"):
    """
    Dynamically create a class with attributes set to data based on the outermost keys of the given dictionary.
    """
    class_attrs = {key: val for key, val in instance_dict.items()}
    return type(class_name, (object,), class_attrs)


def fetch_biomodel_by_term(term: str, index: int = 0):
    """Search for models matching the term and return an instantiated model from BioModels.

        Args:
            term:`str`: search term
            index:`int`: selector index for model choice

        Returns:
            `CDataModel` instance of loaded model.
        TODO: Implement a dynamic search of this
    """
    models = biomodels.search_for_model(term)
    model = models[index]
    return fetch_biomodel(model['id'])


def fetch_biomodel(model_id: str):
    # TODO: make this generalizable for those other than basico
    sbml = biomodels.get_content_for_model(model_id)
    return load_model_from_string(sbml)


def play_composition(instance: dict, duration: int):
    """Configure and run a Composite workflow."""
    workflow = Composite({
        'state': instance
    })
    workflow.run(duration)
    results = workflow.gather_results()
    print(f'RESULTS: {pf(results)}')
    return results


def generate_copasi_process_emitter_schema():
    return generate_emitter_schema(
        emitter_address='local',
        emitter_type='ram-emitter',
        floating_species='tree[float]',
        time='float'
    )


def generate_single_copasi_process_instance(instance_name: str, config: Dict, add_emitter: bool = True) -> Dict:
    """Generate an instance of a single copasi process which is named `instance_name`
        and configured by `config` formatted to the process bigraph Composite API.

        Args:
            instance_name:`str`: name of the new instance referenced by PBG.
            config:`Dict`: see `biosimulators_processes.processes.copasi_process.CopasiProcess`
            add_emitter:`bool`: Adds emitter schema configured for CopasiProcess IO store if `True`. Defaults
                to `True`.
    """
    instance = {}
    instance[instance_name] = {
        '_type': 'process',
        'address': 'local:copasi',
        'config': config,
        'inputs': {
            'floating_species': ['floating_species_store'],
            'model_parameters': ['model_parameters_store'],
            'time': ['time_store'],
            'reactions': ['reactions_store']
        },
        'outputs': {
            'floating_species': ['floating_species_store'],
            'time': ['time_store'],
        }
    }
    if add_emitter:
        instance['emitter'] = generate_copasi_process_emitter_schema()

    return instance


def generate_single_process_instance(
        instance_name: str,
        instance_config: dict,
        inputs_config: dict,
        outputs_config: dict,
        instance_type: str = 'step',
        instance_address: str = 'local:copasi',
) -> dict:
    return {
        instance_name: {
            '_type': instance_type,
            'address': instance_address,
            'config': instance_config,
            'inputs': inputs_config,
            'outputs': outputs_config
        }
    }


def generate_emitter_schema(
        emitter_address: str = "local",
        emitter_type: str = "ram-emitter",
        **emit_values_schema
) -> Dict:
    """
        Args:
            **emit_values_schema:`kwargs`: values to be emitted by the address


    """
    return {
        '_type': 'step',
        'address': 'local:ram-emitter',
        'config': {
            'emit': {**emit_values_schema},
        },
        'inputs': {  # TODO: make this generalized
            'floating_species': ['floating_species_store'],
            'time': ['time_store'],
        }
    }


def generate_parameter_scan_instance(
        num_iterations: int,
        entry_config: Dict,
        modeler: str,
        add_emitter: bool = True,
        *parameters,
        **io_config
) -> Dict:
    """Generate a parameter scan instance configuration for a given process."""
    instance = {}
    for n in range(num_iterations):
        iteration_name = f'{modeler}_{n}'
        iteration_instance = generate_single_process_instance(
            instance_name=iteration_name,
            instance_config=entry_config,
            inputs_config=io_config['inputs'],
            outputs_config=io_config['outputs'])

        for instance_name, instance_schema in iteration_instance:
            instance[instance_name] = iteration_instance

    if add_emitter:
        instance['emitter'] = generate_emitter_schema(
            floating_species='tree[float]',
            time='float')

    return instance


def generate_copasi_parameter_scan_instance(
        num_iterations: int,
        entry_config: Dict,
        # *parameters
):
    # TODO: select parameters
    parameter_scan_instance = {}
    origin_value = 3.00
    for n in range(num_iterations):
        iteration_model_config = generate_sed_model_config_schema(
            entrypoint={'biomodel_id': 'BIOMD0000000051'},
            species_changes={'Extracellular Glucose': {'initial_concentration': origin_value**n}},
            parameter_changes={'catp': {'initial_value': (origin_value - n)**n}}
        )
        iteration_instance = generate_single_copasi_process_instance(
            instance_name=f'copasi_{n}',
            config=iteration_model_config,
            add_emitter=False
        )
        for iter_name, iter_config in iteration_instance.items():
            parameter_scan_instance[iter_name] = iteration_instance

    emitter_schema = generate_emitter_schema(floating_species='tree[float]', time='float')
    parameter_scan_instance['emitter'] = emitter_schema

    return parameter_scan_instance



def generate_sed_model_config_schema(
        entrypoint: Dict,
        species_changes: Dict,
        parameter_changes: Dict,
        reaction_changes: Dict = None
) -> Dict:
    """
        Args:
            entrypoint:`Dict[str, str]`: per CopasiProcess config_schema; ie: {'biomodel_id': 'BIOMODEL>>>>'}
            species_changes:`Dict[str, Dict[str, any]]`: ie: {'a': {'initial_concentration': 32.3}}
            parameter_changes:`Dict[str, Dict[str, any]]`: ie: {'x2': {'expression': '1 -> A..'}}
            reaction_changes:`Dict[str, Union[str, Dict[str, Dict[str, any]]]`:
                    reaction_changes = {
                        'R1': {
                            'reaction_parameters': {
                                '(R1).k1': 23.2
                            },
                            'reaction_scheme': 'A -> B'
                        }

        Example:
            sed_model = sed_schema(
                entrypoint={'biomodel_id': 'BIOMD0000000051'},
                species_changes={'Extracellular Glucose': {'initial_concentration': 5.00}},
                parameter_changes={'catp': {'initial_value': 100.00}},
                reaction_changes={'Aldolase': {'scheme': 'A -> B'}}
            )
    """
    instance_schema = {
        'model': {
            'model_changes': {
                'species_changes': species_changes,
                'parameter_changes': parameter_changes,
                'reaction_changes': reaction_changes
            }
        }
    }

    for param_name, param_val in entrypoint.items():
        instance_schema['model'][param_name] = {
            param_name: param_val
        }

    return instance_schema


def perturb_parameter(num_iterations: int, degree: float):
    _range = []
    for n in range(num_iterations):
        if n > 0:
            n = n * degree
        _range.append(n)
    return _range


def perturb_parameter_numpy(num_iterations: int, degree: float):
    return np.linspace(start=0, stop=num_iterations, num=degree)


def fix_execution_count(notebook_path):
    with open(notebook_path, 'r', encoding='utf-8') as f:
        nb = nbformat.read(f, as_version=4)

    for cell in nb['cells']:
        if cell['cell_type'] == 'code':
            if 'execution_count' not in cell:
                cell['execution_count'] = None
            print('execution_count' in cell)

    with open(notebook_path, 'w', encoding='utf-8') as f:
        nbformat.write(nb, f)


def fix_notebooks_execution_count():
    for root, _, files in os.walk('../composer-notebooks'):
        for _file in files:
            notebook_path = os.path.join(root, _file)
            fix_execution_count(notebook_path)


def get_copasi_parameter_type_default(
        getter: FunctionType,
        param_name: str,
        copasi_model_object: object,
        parameter_type: str
) -> Union[str, int, float]:
    """Return the default value of a given model parameter specified by `parameter_type`.

        Args:
            getter:`function`: basico callback with which to query the model. For example: `get_species`.
            param_name:`str`: parameter name to use as an argument in the `getter`. For example, if getter=get_species, this would be a species name.
            copasi_model_object:`CDataModel`: instance in memory which you are querying.
            parameter_type:`str`: which type of parameter you wish to get the default of based on the getter and param_name.

        For example:

                `get_copasi_parameter_type_default(get_species, 'T', p.copasi_model_object, 'unit')`
    """
    return getter(param_name, model=copasi_model_object)[[parameter_type]].values.tolist()[0][0]


def get_copasi_species_parameter_default(species_name: str, copasi_model_object: object, parameter_type: str):
    from basico import get_species
    return get_copasi_parameter_type_default(
        getter=get_species,
        param_name=species_name,
        copasi_model_object=copasi_model_object,
        parameter_type=parameter_type)
